/*
 *   $Id: $
 *
 *   Authors:
 *    Greg Daley  <greg.daley@eng.monash.edu.au>  
 *
 *   This software is Copyright 2004 by the above mentioned author(s), 
 *   All Rights Reserved.
 *
 *   The license which is distributed with this software in the file COPYRIGHT
 *   applies to this software. If your distribution is missing this file, you
 *   may request it from <lutchann@litech.org>.
 *
 */

#include <config.h>
#include <includes.h>
#include <radvd.h>
#include <r2router.h>
#include <detfra.h>

extern int sock;

int r2r_handledetfrao(struct Interface *iface, struct nd_opt_detfra
		      *dfra, struct sockaddr_in6 *addr, int ival);

struct DetFRARouter *detfra_list_update(struct Interface *iface,
					uint32_t rankscore,
					struct nd_opt_detfra
					*detfrao, struct in6_addr *addr,
					int ival);

struct DetFRARouter *detfra_list_clean(struct Interface *iface);

int islinkid(struct dna_linkid *linkid)
{
	return (linkid->link_id16[0] || linkid->link_id16[1] ||
		linkid->link_id16[2]);
}

int linkid_compare(struct dna_linkid *orig, struct dna_linkid *targ)
{

	uint64_t a=0, b = 0;
	int i;
	uint64_t modulus = 0x7fffffff;
	modulus <<= 16;
	modulus |= 0xffff;



	a <<= 63;
	b <<= 63;

	for (i = 0; i < 6; i++) {
		a <<= 8;
		b <<= 8;
		a += orig->link_id[i];
		b += targ->link_id[i];
		/*dlog(LOG_DEBUG, 4, "COMPARE a[%d]: %hhx (%llx) b[%d]: %hhx (%llx)",
		   i,  orig->link_id[i], a, i ,  targ->link_id[i],b );   */
	}

/*        dlog(LOG_DEBUG, 4, "COMPARE a: %-20llx b: %-20llx modulus: %llx CL:%d",
                      a, b, modulus,
   ((( a+modulus ) - b )  % modulus)  >
                                        ((( b+modulus ) - a )  % modulus)); */

	return (((a + modulus) - b) % modulus) >
	    (((b + modulus) - a) % modulus);
}

int linkid_add(struct dna_linkid *targ, struct dna_linkid *add)
{

	uint64_t a=0, b=0, s = 0;
	uint64_t modulus = 0x7fffffff;
	int i;
	modulus <<= 16;
	modulus += 0xffff;

	a <<= 63;
	b <<= 63;

	for (i = 0; i < 6; i++) {
		a <<= 8;
		b <<= 8;
		a += targ->link_id[i];
		b += add->link_id[i];
	}

	dlog(LOG_DEBUG, 4,
	     "ADD a: %-20llx b: %-20llx s:%-20llx modulus:%-20llx", a, b,
	     s, modulus);

	s = (a + b % modulus);
	dlog(LOG_DEBUG, 4,
	     "SUM a: %-20llx s:%-20llx modulus:%-20llx CL:%d", a, s,
	     modulus,
	     (((a + modulus) - s) % modulus) >
	     (((s + modulus) - a) % modulus));
	for (i = 5; i >= 0; i--) {
		add->link_id[i] = (uint8_t) (0xff & s);
		s >>= 8;
	}
	dlog(LOG_DEBUG, 4,
	     "ADD a: %-20llx b: %-20llx s:%-20llx modulus:%-20llx"
	     " compare: %d\n", a, b, s, modulus, linkid_compare(targ,
								add));
	memcpy(targ, add, sizeof(struct dna_linkid));

	return 1;
}

int initRandom(uint8_t * buf, int len, int urandom)
{
	int rfd;
	int numread = 0;
	char *randf;
	int tally = 0;

	if (urandom) {
		randf = NONBLOCKING_RNG_NAME;
	} else {
		randf = BLOCKING_RNG_NAME;
	}

	if ((rfd = open(randf, O_RDONLY)) < 0) {
		return -1;
	}
	/* this may take a while on routers with low randomness
	   (no mouse movements...), since a depleted random
	   pool causes blocks on /dev/random */
	while ((numread = read(rfd, buf + tally, len - tally)) > 0) {
		dlog(LOG_DEBUG, 4, "try to get a read from %s nr:%d t:%d",
		     randf, numread, tally);

		tally += numread;
	}
	return 0;

}

void r2router_handler(void *data)
{
        struct timer_lst *tm= (struct timer_lst *)data;
	struct Interface *iface;
	int request = 0;
	int oldrank;
	double next;
	double nextra;
	struct dna_linkid tmplinkid;
	struct timeval tv;

  	iface = tm->iface;

	dlog(LOG_DEBUG, 4, "r2router_handler called for %s (%d c:%d)",
	     iface->Name, iface->r2rState, iface->r2rStateCount);

	oldrank = iface->dfraSelf.CurrRank;

	gettimeofday(&tv, NULL);

	if (iface->DetermineFastRA)
		detfra_list_clean(iface);

	if ((iface->r2rState & R2R_STATES_READY) &&
	    (oldrank != iface->dfraSelf.CurrRank)) {
		dlog(LOG_DEBUG, 4, " dropped entries /changed rank ");
		iface->dfraSelf.LastRank = iface->dfraSelf.CurrRank;
		iface->r2rState = R2R_STATE_ADV_CHANGE;
		iface->r2rStateCount = 0;
		iface->dfraSelf.FixedDelay =
		    iface->dfraFirst.next->Separation *
		    (iface->dfraSelf.CurrRank - 1);
		next = (double) rand_between(0.0, MAX_RTR_STATUS_DELAY);
		/*clear_timer(&iface->rtrtm); */
		set_timer(&iface->rtrtm, next);
		return;
	}
	if (iface->UseLinkID) {
		flog(LOG_INFO,
		     "ProspectiveLinkID %04x%04x%04x Current %04x%04x%04x on %s",
		     ntohs(iface->ProspectiveLinkID.link_id16[0]),
		     ntohs(iface->ProspectiveLinkID.link_id16[1]),
		     ntohs(iface->ProspectiveLinkID.link_id16[2]),
		     ntohs(iface->CurrentLinkID.link_id16[0]),
		     ntohs(iface->CurrentLinkID.link_id16[1]),
		     ntohs(iface->CurrentLinkID.link_id16[2]),
		     iface->Name);
	}
	iface->r2rStateCount++;
	switch (iface->r2rState) {
	case R2R_STATE_BOOT_STRAP:
		next = (double) RTR_STATUS_REQ_INTERVAL;
		request = 1;
		send_r2r(sock, iface, NULL, request);
		if (iface->r2rStateCount == MAX_INITIAL_RTR_STATUS) {
			iface->r2rState = R2R_STATE_ADV_CHANGE;
			iface->r2rStateCount = 0;

			if (iface->DetermineFastRA) {
				iface->dfraSelf.LastRank =
				    iface->dfraSelf.CurrRank;
				iface->dfraSelf.FixedDelay =
				    iface->dfraFirst.next->Separation *
				    (iface->dfraSelf.CurrRank - 1);

			}
			next = MIN_DELAY_BETWEEN_RTR_STATUS +
			    (double) rand_between(0.0,
						  MAX_RTR_STATUS_DELAY);

		}

		break;
	case R2R_STATE_ADV_CHANGE:
		next = (double) MIN_DELAY_BETWEEN_RTR_STATUS;
		/* only happens on entering state first time */
		if (iface->UseLinkID) {
			if (iface->ForceLinkIDChange &&
			    islinkid(&iface->ProspectiveLinkID)) {
				flog(LOG_ERR,
				     "Force LinkID change on %s B%d",
				     iface->Name,
				     iface->LinkIDBrokenRandom);

				initRandom((uint8_t *) & tmplinkid,
					   sizeof(struct dna_linkid),
					   iface->LinkIDBrokenRandom);

				/* copy prospectiveLinkID to current.
				   Add a number bigger than current in
				   same half */
				tmplinkid.link_id[0] &= 0x8f;
				linkid_add(&iface->ProspectiveLinkID,
					   &tmplinkid);
				memcpy(&iface->LastLinkIdChange,&tv, sizeof(
						struct timeval));
			} else if (!islinkid(&iface->CurrentLinkID) &&
				   !islinkid(&iface->ProspectiveLinkID)) {
				/* no preexisting LinkID */
				flog(LOG_INFO,
				     "LinkID init on %s B%d",
				     iface->Name,
				     iface->LinkIDBrokenRandom);
				initRandom((uint8_t *)
					   & iface->ProspectiveLinkID,
					   sizeof(struct dna_linkid),
					   iface->LinkIDBrokenRandom);
				flog(LOG_INFO,
				     "ProspectiveLinkID %04x%04x%04x on %s",
				     ntohs(iface->ProspectiveLinkID.
					   link_id16[0]),
				     ntohs(iface->ProspectiveLinkID.
					   link_id16[1]),
				     ntohs(iface->ProspectiveLinkID.
					   link_id16[2]), iface->Name);
				memcpy(&iface->LastLinkIdChange,&tv, sizeof(
						struct timeval));
			}
			/* have linkid */
			iface->ForceLinkIDChange = 0;
		}
		if (iface->r2rStateCount == MAX_INITIAL_RTR_STATUS) {
			iface->r2rState = R2R_STATE_STEADY;
			iface->r2rStateCount = 0;
			if (iface->UseLinkID) {
				memcpy(&iface->CurrentLinkID,
				       &iface->ProspectiveLinkID,
				       sizeof(struct dna_linkid));
			}
		}
		send_r2r(sock, iface, NULL, request);
		/* schedule nearest Multicast RA */
		clear_timer(&iface->tm);
		if ((nextra = (((double)MIN_DELAY_BETWEEN_RAS) -
			       tv_diff(&tv, &iface->last_multicast))) < 0) {
			memcpy(&iface->last_multicast, &tv,
					sizeof(struct timeval));
			send_ra(sock, iface, NULL, NULL);
			nextra = rand_between(iface->MinRtrAdvInterval,
					      iface->MaxRtrAdvInterval);
		}
		set_timer(&iface->tm, nextra);
		break;
	case R2R_STATE_STEADY:
		next = (double) iface->RouterToRouterStatusInt;
		if ((iface->r2rStateCount % 100) == 0) {
			/* Add some more randomness each 5 minutes */
			next += (double) rand_between(0.0,
						      MAX_RTR_STATUS_DELAY);
		}
		send_r2r(sock, iface, NULL, request);
		break;
	default:
		flog(LOG_ERR, "Unknown RouterToRouter state on %s (%d)",
		     iface->Name, iface->r2rState);
		return;
		break;
	}

	if (iface->UseLinkID) {
		flog(LOG_INFO,
		     "ProspectiveLinkID %04x%04x%04x Current %04x%04x%04x on %s",
		     ntohs(iface->ProspectiveLinkID.link_id16[0]),
		     ntohs(iface->ProspectiveLinkID.link_id16[1]),
		     ntohs(iface->ProspectiveLinkID.link_id16[2]),
		     ntohs(iface->CurrentLinkID.link_id16[0]),
		     ntohs(iface->CurrentLinkID.link_id16[1]),
		     ntohs(iface->CurrentLinkID.link_id16[2]),
		     iface->Name);
	}
	set_timer(&iface->rtrtm, next);
}


void r2rtimers(struct Interface *iface)
{
	double next;
	init_timer(&iface->rtrtm, r2router_handler, (void *) iface);
	next = rand_between(0.0, 1.0);
	set_timer(&iface->rtrtm, next);
}

uint32_t get_detfra_score(uint32_t relpref, struct in6_addr *lladdr)
{
	uint32_t tmp;
	tmp = ((ntohl(lladdr->s6_addr32[3]) & 0x00ffffff) ^ 0x00ffffff);
	return tmp | (relpref << 24);
}


void
init_detfrouter(struct DetFRARouter *rtr, uint32_t relpref,
		struct nd_opt_detfra *detfrao, struct in6_addr *addr)
{
	memcpy(&rtr->lladdr, addr, sizeof(struct in6_addr));
	rtr->RankScore = get_detfra_score(relpref, addr);
	rtr->next = rtr->prev = NULL;

	if (!detfrao) {
		rtr->LastRank = ND_OPT_DETFRA_BOOTSTRAP_RANK;
		rtr->CurrRank = ND_OPT_DETFRA_BOOTSTRAP_RANK;
		rtr->Separation = DFLT_FastRASeparation;
		rtr->FixedDelay = 500;
		rtr->RelPref = ND_OPT_DETFRA_DEF_REL_PREF;
		rtr->LastUpdate = 0xff000000;
	} else {
		rtr->LastRank = detfrao->nd_opt_detfra_rank;
		rtr->CurrRank = detfrao->nd_opt_detfra_rank;
		rtr->RelPref = detfrao->nd_opt_detfra_relpref;
		rtr->Separation = detfrao->nd_opt_detfra_separation;
		rtr->FixedDelay = ntohs(detfrao->nd_opt_detfra_fixeddelay);
		rtr->LastUpdate = 0xff000000;
	}
}

void iface_DetFastRA_init(struct Interface *iface)
{
	uint8_t unspec[16] =
	    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	uint8_t allones[16] =
	    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xff, 0xff, 0xff, 0xff };

	char addr_str[INET6_ADDRSTRLEN];

	print_addr(&iface->if_addr, addr_str);


	dlog(LOG_DEBUG, 4,
	     "Initializing Deterministic FastRA with addr: %s", addr_str);
	init_detfrouter(&iface->dfraFirst, 0, 0,
			(struct in6_addr *) &allones);
	init_detfrouter(&iface->dfraSelf, iface->DetFastRARelPref, 0,
			&iface->if_addr);
	init_detfrouter(&iface->dfraLast, 255, 0,
			(struct in6_addr *) &unspec);


	iface->dfraSelf.RelPref = iface->DetFastRARelPref;
	iface->dfraSelf.Separation = iface->FastRASeparation;

	/* Put self in the middle */
	iface->dfraLast.next = &iface->dfraLast;
	iface->dfraLast.prev = &iface->dfraSelf;
	iface->dfraSelf.next = &iface->dfraLast;
	iface->dfraSelf.prev = &iface->dfraFirst;
	iface->dfraFirst.next = &iface->dfraSelf;
	iface->dfraFirst.prev = &iface->dfraFirst;
	iface->dfraSelf.CurrRank = 1;	/* before we hear from anyone, we lead */
}


int r2r_handledetfrao(struct Interface *iface, struct nd_opt_detfra
		      *detfrao, struct sockaddr_in6 *addr, int ival)
{
	uint32_t rankscore;
	uint8_t oldrank;
	double next;

	rankscore = get_detfra_score(detfrao->nd_opt_detfra_relpref,
				     &addr->sin6_addr);

	dlog(LOG_DEBUG, 4,
	     "Type:%-3d Rank:%-5x Pref:%x Sep:%x FixDelay:%u Score:%x i:%d",
	     detfrao->nd_opt_detfra_type,
	     detfrao->nd_opt_detfra_rank,
	     detfrao->nd_opt_detfra_relpref,
	     detfrao->nd_opt_detfra_separation,
	     ntohs(detfrao->nd_opt_detfra_fixeddelay), rankscore, ival);

	oldrank = iface->dfraSelf.CurrRank;

	detfra_list_clean(iface);

	detfra_list_update(iface, rankscore, detfrao, &addr->sin6_addr,
			   ival);

	if ((iface->r2rState & R2R_STATES_READY) &&
	    (oldrank != iface->dfraSelf.CurrRank)) {
		dlog(LOG_DEBUG, 4, "r2r_handledetfrao changed rank");
		iface->dfraSelf.LastRank = iface->dfraSelf.CurrRank;
		iface->r2rState = R2R_STATE_ADV_CHANGE;
		iface->r2rStateCount = 0;
		iface->dfraSelf.FixedDelay =
		    iface->dfraFirst.next->Separation *
		    (iface->dfraSelf.CurrRank - 1);
		next = (double) rand_between(0.0, MAX_RTR_STATUS_DELAY);
		clear_timer(&iface->rtrtm);
		set_timer(&iface->rtrtm, next);
	}
	return 0;

}

int r2r_handlelinkid(struct Interface *iface, struct nd_opt_linkid *lid)
{
	double next;

	if (!islinkid(&lid->nd_opt_linkid_id))
		return -1;

	if (!islinkid(&iface->CurrentLinkID) &&
	    !islinkid(&iface->ProspectiveLinkID)) {
		dlog(LOG_DEBUG, 4, "Set LinkID");
		/* At least set prospective */
		memcpy(&iface->ProspectiveLinkID, &lid->nd_opt_linkid_id,
		       sizeof(struct dna_linkid));
		return 1;
	}
	dlog(LOG_DEBUG, 4, "Handle LinkID"
	     "R %04x%04x%04x P %04x%04x%04x",
	     ntohs(lid->nd_opt_linkid_id.link_id16[0]),
	     ntohs(lid->nd_opt_linkid_id.link_id16[1]),
	     ntohs(lid->nd_opt_linkid_id.link_id16[2]),
	     ntohs(iface->ProspectiveLinkID.link_id16[0]),
	     ntohs(iface->ProspectiveLinkID.link_id16[1]),
	     ntohs(iface->ProspectiveLinkID.link_id16[2]));
	if (linkid_compare
	    (&iface->ProspectiveLinkID, &lid->nd_opt_linkid_id) > 0) {
		/* received is better (modulo 2^48 -1 )    */
		/* assumes Prospective LinkID is either */
		/* Same as Current or better */
		dlog(LOG_DEBUG, 4, "Better LinkID");

		memcpy(&iface->ProspectiveLinkID, &lid->nd_opt_linkid_id,
		       sizeof(struct dna_linkid));

		if (iface->r2rState & R2R_STATES_READY) {
			/* rename variable ! */
			iface->r2rState = R2R_STATE_ADV_CHANGE;
			iface->r2rStateCount = 0;
			next = (double) rand_between(0.0,
						     MAX_RTR_STATUS_DELAY);
			clear_timer(&iface->rtrtm);
			set_timer(&iface->rtrtm, next);
		}

		return 1;
	}
	return 0;
}

struct DetFRARouter *detfra_list_clean(struct Interface *iface)
{
	struct DetFRARouter *tmp  =  NULL;
	struct DetFRARouter *old;

	int position = 1;
	char addr_str[INET6_ADDRSTRLEN];
	time_t secs;

	print_addr(&iface->if_addr, addr_str);

	secs = time(NULL);
	dlog(LOG_DEBUG, 4, "detfra_list_clean clearing interface %s ",
	     addr_str);
	tmp = iface->dfraFirst.next;

	while (tmp != &iface->dfraLast) {
		print_addr(&tmp->lladdr, addr_str);
		dlog(LOG_DEBUG, 5, "clearing entry %i %s", position,
		     addr_str);
		if ((tmp != &iface->dfraSelf)
		    &&
		    (((uint32_t) tmp->LastUpdate +
		      (uint32_t) tmp->twoival) < (uint32_t) secs)) {
			/* rough granularity */
			print_addr(&tmp->lladdr, addr_str);
			dlog(LOG_DEBUG, 4,
			     "detfra_list_clean remove: %d %s %lu i%lu t:%lu",
			     position, addr_str, tmp->LastUpdate +
			     tmp->twoival, tmp->twoival, secs);
			old = tmp;
			tmp = tmp->next;
			old->prev->next = old->next;
			old->next->prev = old->prev;
			old->next = old->prev = NULL;
			continue;
		}
		tmp->CurrRank = position++;
		tmp = tmp->next;
	}

	return iface->dfraFirst.next;
}

struct DetFRARouter *detfra_list_update(struct Interface *iface,
					uint32_t rankscore,
					struct nd_opt_detfra
					*detfrao, struct in6_addr *addr,
					int ival)
{
	struct DetFRARouter *tmp, *nextemp = NULL;
	int position = 1;
	char addr_str[INET6_ADDRSTRLEN];
	time_t secs;

	print_addr(addr, addr_str);
	secs = time(NULL);

	tmp = iface->dfraFirst.next;

	/* currently assuming that no-one is sending changed 
	   rankscore entries */
	dlog(LOG_DEBUG, 4, "detfra_list_update Target: %lx   %s %i",
	     rankscore, addr_str, ival);


	while (tmp != &iface->dfraLast) {

		print_addr(&tmp->lladdr, addr_str);
		if ((tmp->RankScore == rankscore) &&
		    !memcmp(&tmp->lladdr, addr, sizeof(struct in6_addr))) {
			dlog(LOG_DEBUG, 5,
			     "DFRAList match P:%d m%x from %s", position,
			     tmp->RankScore, addr_str);
			/* same entry, so update */
			/* make sure no-one ends our address here */
			tmp->LastRank = detfrao->nd_opt_detfra_rank;
			tmp->CurrRank = position++;
			tmp->twoival = (time_t) (((2 * (unsigned) ival) + 999) / 1000);	/* secs */
			tmp->LastUpdate = secs;
			nextemp = tmp;
			tmp = tmp->next;
			break;
		}

		if (tmp->RankScore > rankscore) {
			/* insert before worse */
			break;
		}
		dlog(LOG_DEBUG, 5, "DFRAList P:%d (m%lx) from %s",
		     position, tmp->RankScore, addr_str);

		tmp->CurrRank = position++;
		tmp = tmp->next;
	}
	if (!nextemp) {
		print_addr(addr, addr_str);
		dlog(LOG_DEBUG, 4, "DFRAList insert P:%d (m%lx) from %s",
		     position, rankscore, addr_str);
		if ((nextemp = (struct DetFRARouter *)
		     malloc(sizeof(struct DetFRARouter))) == NULL) {
			return NULL;
		}

		init_detfrouter(nextemp, detfrao->nd_opt_detfra_relpref,
				detfrao, addr);
		nextemp->CurrRank = position++;
		tmp->prev->next = nextemp;
		nextemp->prev = tmp->prev;
		tmp->prev = nextemp;
		tmp->twoival = (time_t) (((2 * ival) + 999) / 1000);	/* secs */
		tmp->LastUpdate = secs;
		nextemp->next = tmp;
	}
	while (tmp != &iface->dfraLast) {
		print_addr(&tmp->lladdr, addr_str);
		dlog(LOG_DEBUG, 4, "DFRAList P:%d (m%lx) from %s",
		     position, tmp->RankScore, addr_str);
		tmp->CurrRank = position++;
		tmp = tmp->next;
	}
	return nextemp;
}
void iface_LinkID_init(struct Interface *iface)
{

	char addr_str[INET6_ADDRSTRLEN];
	print_addr(&iface->if_addr, addr_str);


	dlog(LOG_DEBUG, 4, "Initializing LinkID with addr: %s", addr_str);
	memset(&iface->CurrentLinkID, 0, sizeof(struct dna_linkid));
	memset(&iface->ProspectiveLinkID, 0, sizeof(struct dna_linkid));

	/* set state */
}

       /* draft-daley-dna-det-fastra-00.txt Extension */

 /* switch on router-to-router processing on Interface */

int check_routerconf(struct Interface *iface)
{
	int res = 0;
	flog(LOG_INFO, "routerconf:   DFRA: %d   LINKID %d",
	     iface->DetermineFastRA, iface->UseLinkID);

	if (iface->DetermineFastRA) {
		iface->RouterToRouter = 1;
		iface_DetFastRA_init(iface);
	}
	if (iface->UseLinkID) {
		iface->RouterToRouter = 1;
		iface_LinkID_init(iface);
	}
	if (iface->RouterToRouter) {
		iface->r2rState = R2R_STATE_BOOT_STRAP;
		iface->r2rStateCount = 0;
		/* timer callback scheduling */
		flog(LOG_INFO, "using RouterToRouter messages");
		if (iface->RouterToRouterStatusInt <
		    MIN_DELAY_BETWEEN_RTR_STATUS) {
			/* currently don't care about maximum */
			flog(LOG_INFO,
			     "Invalid RouterToRouterStatusInt < 3 on %s",
			     iface->Name);
			exit(1);
		}
	}
	return res;
}
